// Copyright 2000-2024 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
package com.intellij.codeInsight.inline.completion.logs

import com.intellij.codeInsight.inline.completion.InlineCompletionEventAdapter
import com.intellij.codeInsight.inline.completion.InlineCompletionEventType
import com.intellij.codeInsight.inline.completion.logs.FinishingLogs.FINISH_TYPE
import com.intellij.codeInsight.inline.completion.logs.FinishingLogs.TIME_TO_START_SHOWING
import com.intellij.codeInsight.inline.completion.logs.FinishingLogs.WAS_SHOWN
import com.intellij.codeInsight.inline.completion.logs.InlineCompletionLogsContainer.Phase
import com.intellij.codeInsight.inline.completion.logs.StartingLogs.FILE_LANGUAGE
import com.intellij.codeInsight.inline.completion.logs.StartingLogs.INLINE_API_PROVIDER
import com.intellij.codeInsight.inline.completion.logs.StartingLogs.REQUEST_EVENT
import com.intellij.codeInsight.inline.completion.logs.StartingLogs.REQUEST_ID
import com.intellij.internal.statistic.eventLog.events.EventFields
import com.intellij.openapi.application.readAction
import com.intellij.openapi.editor.Editor
import java.util.concurrent.atomic.AtomicBoolean
import java.util.concurrent.atomic.AtomicLong

internal class InlineCompletionLogsListener(private val editor: Editor) : InlineCompletionEventAdapter {
  private val lastInvocationTimestamp = AtomicLong()
  private val wasShown = AtomicBoolean()

  override fun onRequest(event: InlineCompletionEventType.Request) {
    lastInvocationTimestamp.set(event.lastInvocation)
    wasShown.set(false)
    val container = InlineCompletionLogsContainer.create(event.request.editor)
    container.add(REQUEST_ID with event.request.requestId)
    container.add(REQUEST_EVENT with event.request.event.javaClass)
    container.add(INLINE_API_PROVIDER with event.provider)
    event.request.event.toRequest()?.file?.language?.let { container.add(FILE_LANGUAGE with it) }
    container.addAsync {
      readAction {
        InlineCompletionContextLogs.getFor(event.request)
      }
    }
  }

  override fun onShow(event: InlineCompletionEventType.Show) {
    if (wasShown.getAndSet(true)) return
    val container = InlineCompletionLogsContainer.get(editor) ?: return
    container.add(TIME_TO_START_SHOWING with (System.currentTimeMillis() - lastInvocationTimestamp.get()))
  }

  override fun onHide(event: InlineCompletionEventType.Hide) {
    val container = InlineCompletionLogsContainer.remove(editor) ?: return
    container.add(WAS_SHOWN with wasShown.get())
    container.add(FINISH_TYPE with event.finishType)
    container.logCurrent() // see doc of this function, it's very fast, and we should wait for its completion
  }
}

private object StartingLogs : PhasedLogs(Phase.INLINE_API_STARTING) {
  val REQUEST_ID = register(EventFields.Long("request_id"))
  val REQUEST_EVENT = register(EventFields.Class("request_event"))
  val INLINE_API_PROVIDER = register(EventFields.Class("inline_api_provider"))
  val FILE_LANGUAGE = register(EventFields.Language("file_language"))
}

private object FinishingLogs : PhasedLogs(Phase.INLINE_API_FINISHING) {
  val WAS_SHOWN = register(EventFields.Boolean("was_shown"))
  val TIME_TO_START_SHOWING = register(EventFields.Long("time_to_start_showing"))
  val FINISH_TYPE = register(EventFields.Enum("finish_type", InlineCompletionUsageTracker.ShownEvents.FinishType::class.java))
}

internal class InlineCompletionListenerSessionLogs : InlineCompletionSessionLogsEP {
  override val fields = listOf(StartingLogs, FinishingLogs)
}